﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     main.cpp

  Copyright (C)2009-2013 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 50863 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/fnd.h>
#include <nn/fs.h>
#include <nn/snd.h>
#include <nn/applet.h>
#include <nn/aacdec.h>

#include "demo.h"
#include "applet.h"

extern const u8 DSPAUDIO_BEGIN[];

namespace
{
    const size_t HEAP_SIZE_FOR_GX = 8 * 1024 * 1024;
    demo::RenderSystemDrawing s_RenderSystem;

    nn::fnd::ExpHeap    s_ExpHeap;
    const size_t        DEVICE_MEMORY_SIZE = (32 * 1024 * 1024);
    const char          FILE_NAME[] = "rom:/kart_title.32.aac";

    const s32           DECODER_WAVE_BUFFER_NUM = 4;
    s16*                s_pStreamBuffer[DECODER_WAVE_BUFFER_NUM];
    nn::snd::WaveBuffer s_StreamWaveBuffer[DECODER_WAVE_BUFFER_NUM];

    u8*                 s_pBitstreamBuffer;
    u8*                 s_pBitstreamInputBuffer;
    s32                 s_BitstreamFileSize;
    s32                 s_BitstreamPos = 0;
    const size_t        DECODED_PCM_BUFFER_SIZE = NN_DSP_CODEC_AAC_PCM_SAMPLES * sizeof(s16) * 2;

    s16*                s_pDecodedPcmBufferL;
    s16*                s_pDecodedPcmBufferR;

    void*               s_pStateBuffer;

    void Decode(s16* pOutput)
    {
        // Parse header and extract frame.
        u8* p = s_pBitstreamBuffer + s_BitstreamPos;
        s32 frameSize = nn::aacdec::GetAdtsFrameSize(p);
        if (frameSize > 0)
        {
            ::std::memcpy(s_pBitstreamInputBuffer, p, frameSize);
            nn::dsp::FlushDataCache(reinterpret_cast<uptr>(s_pBitstreamInputBuffer), frameSize);

            // Wait until decoding and processing complete.
            nn::aacdec::Decode(s_pBitstreamInputBuffer, s_pDecodedPcmBufferL, s_pDecodedPcmBufferR);
            nn::aacdec::Wait();
            // Verify result
            if (nn::aacdec::GetErrorCode() != 0)
            {
                NN_PANIC("Failed to decode!\n");
            }
            s_BitstreamPos += nn::aacdec::GetUsedBytes();

            for (int i = 0; i < NN_DSP_CODEC_AAC_PCM_SAMPLES; i++)
            {
                pOutput[2*i+0] = s_pDecodedPcmBufferL[i];
                pOutput[2*i+1] = s_pDecodedPcmBufferR[i];
            }
            nn::dsp::FlushDataCache(reinterpret_cast<uptr>(pOutput), DECODED_PCM_BUFFER_SIZE);

            // After we read to the end of the file, return to the start.
            if (s_BitstreamPos >= s_BitstreamFileSize)
            {
                s_BitstreamPos = 0;
            }
        }
        else
        {
            NN_PANIC("Invalid aac file!\n");
        }
    }

    const s32           DECODE_THREAD_PRIORITY = 4;
    const size_t        DECODE_THREAD_STACK_SIZE = 4096;
    nn::os::Thread      s_DecodeThread;
    nn::os::Event       s_DecodeEvent;
    nn::os::Event       s_DecodeThreadExitEvent;
    s32                 s_DecodeBufferId;
    enum
    {
        EVENT_DECODE,
        EVENT_EXIT,
        EVENT_NUM
    };
    void DecodeThreadFunc(uptr arg)
    {
        NN_UNUSED_VAR(arg);

        s_DecodeEvent.Initialize(false);
        s_DecodeThreadExitEvent.Initialize(false);

        nn::os::Event *events[EVENT_NUM] = { &s_DecodeEvent, &s_DecodeThreadExitEvent };

        while (true)
        {
            s32 num = nn::os::WaitObject::WaitAny(reinterpret_cast<nn::os::WaitObject**>(events), EVENT_NUM);

            if (num == EVENT_DECODE)
            {
                Decode(s_pStreamBuffer[s_DecodeBufferId]);
            }
            if (num == EVENT_EXIT)
            {
                break;
            }
        }
    }

    // Code for thread awaiting interrupt notification
    const s32           SOUND_THREAD_PRIORITY = 2;
    const size_t        SOUND_THREAD_STACK_SIZE = 4096;
    nn::os::Thread      s_SoundThread;
    bool                s_IsSoundThreadEnabled;
    void SoundThreadFunc(uptr arg)
    {
        NN_UNUSED_VAR(arg);

        // Initialize DSP and SND libraries
        NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
        NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadComponent(DSPAUDIO_BEGIN, NN_DSPCOMPONENT_SIZE));
        NN_UTIL_PANIC_IF_FAILED(nn::snd::Initialize());

        // Load AAC file
        nn::fs::FileReader reader(FILE_NAME);
        s_BitstreamFileSize = reader.GetSize();
        s_pBitstreamBuffer = reinterpret_cast<u8*>(s_ExpHeap.Allocate(s_BitstreamFileSize));
        NN_ASSERT(s_pBitstreamBuffer);
        reader.Read(s_pBitstreamBuffer, s_BitstreamFileSize);
        reader.Finalize();

        // Allocate stream playback buffer (must be 32-byte aligned).
        for (int i = 0; i < DECODER_WAVE_BUFFER_NUM; i++)
        {
            s_pStreamBuffer[i] = reinterpret_cast<s16*>(s_ExpHeap.Allocate(DECODED_PCM_BUFFER_SIZE, 32));
            NN_ASSERT(s_pStreamBuffer[i]);
        }

        // Allocate PCM buffer and bitstream buffer for decoding (must be 32-byte aligned).
        s_pBitstreamInputBuffer = reinterpret_cast<u8*>(s_ExpHeap.Allocate(NN_DSP_CODEC_AAC_BS_SIZE, 32));
        NN_ASSERT(s_pBitstreamInputBuffer);
        s_pDecodedPcmBufferL = reinterpret_cast<s16*>(s_ExpHeap.Allocate(DECODED_PCM_BUFFER_SIZE/2, 32));
        NN_ASSERT(s_pDecodedPcmBufferL);
        s_pDecodedPcmBufferR = reinterpret_cast<s16*>(s_ExpHeap.Allocate(DECODED_PCM_BUFFER_SIZE/2, 32));
        NN_ASSERT(s_pDecodedPcmBufferR);

        // Settings for buffer that backs up the decoder’s internal state when the system sleeps.
        s_pStateBuffer = s_ExpHeap.Allocate(NN_AACDEC_STATE_BUFFER_SIZE, 32);
        NN_ASSERT(s_pStateBuffer);

        // Initialize decoder
        if (nn::aacdec::Initialize(nn::aacdec::FORM_ADTS) == false)
        {
            NN_PANIC("error: init\n");
        }

        // Settings for buffer that backs up the decoder’s internal state when the system sleeps.
        // If these settings are not configured, the decoder’s state is initialized when the system sleeps.
        nn::aacdec::SetStateBuffer(s_pStateBuffer);

        // Decode first DECODER_WAVE_BUFFER_NUM frame
        for (int i = 0; i < DECODER_WAVE_BUFFER_NUM; i++)
        {
            Decode(s_pStreamBuffer[i]);
        }

        // Check number of channels (this sample demo only supports stereo).
        if (nn::aacdec::GetChannelConfig() != 2)
        {
            NN_PANIC("This sample demo supports only stereo streams\n");
        }

        // Prepare Voice
        nn::snd::Voice* pVoice = nn::snd::AllocVoice(128, NULL, NULL);
        NN_ASSERT(pVoice);
        pVoice->SetChannelCount(nn::aacdec::GetChannelConfig());
        pVoice->SetSampleFormat(nn::snd::SAMPLE_FORMAT_PCM16);
        pVoice->SetSampleRate(nn::aacdec::GetSampleRate());
        pVoice->SetPitch(1.0f);
        nn::snd::MixParam mix;
        mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.707f;
        mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.707f;
        pVoice->SetMixParam(mix);
        pVoice->SetVolume(1.0f);

        // Initialize and add buffer
        for (int i = 0; i < DECODER_WAVE_BUFFER_NUM; i++)
        {
            nn::snd::InitializeWaveBuffer(&s_StreamWaveBuffer[i]);
            s_StreamWaveBuffer[i].bufferAddress = s_pStreamBuffer[i];
            s_StreamWaveBuffer[i].sampleLength =
                nn::snd::GetSampleLength(DECODED_PCM_BUFFER_SIZE, nn::snd::SAMPLE_FORMAT_PCM16, 2);
            s_StreamWaveBuffer[i].loopFlag = false;
            pVoice->AppendWaveBuffer(&s_StreamWaveBuffer[i]);
        }

        // Start playback
        pVoice->SetState(nn::snd::Voice::STATE_PLAY);

        int id = 0;
        s_IsSoundThreadEnabled = true;
        while (s_IsSoundThreadEnabled)
        {
            nn::snd::WaitForDspSync();  // Wait for data from DSP

            // Start decoder if necessary.
            if (pVoice->GetState() == nn::snd::Voice::STATE_PLAY &&
                s_StreamWaveBuffer[id].status == nn::snd::WaveBuffer::STATUS_DONE)
            {
                s_DecodeBufferId = id;
                nn::snd::WaveBuffer* pBuf = &s_StreamWaveBuffer[s_DecodeBufferId];

                nn::snd::InitializeWaveBuffer(pBuf);
                pBuf->bufferAddress = s_pStreamBuffer[s_DecodeBufferId];
                pBuf->sampleLength = 
                    nn::snd::GetSampleLength(DECODED_PCM_BUFFER_SIZE, nn::snd::SAMPLE_FORMAT_PCM16, 2);
                pBuf->loopFlag = false;
                pVoice->AppendWaveBuffer(pBuf);  // Add buffer

                s_DecodeEvent.Signal();
                if (++id >= DECODER_WAVE_BUFFER_NUM)
                {
                    id = 0;
                }
            }

            nn::snd::SendParameterToDsp();  // Send arguments to DSP
        }

        // Stops sound.
        pVoice->SetState(nn::snd::Voice::STATE_STOP);
        nn::snd::FreeVoice(pVoice);

        // Finalize aacdec.
        nn::aacdec::Finalize();

        // Memory deallocation
        for (int i = 0; i < DECODER_WAVE_BUFFER_NUM; i++)
        {
            s_ExpHeap.Free(s_pStreamBuffer[i]);
        }
        s_ExpHeap.Free(s_pBitstreamBuffer);
        s_ExpHeap.Free(s_pBitstreamInputBuffer);
        s_ExpHeap.Free(s_pDecodedPcmBufferL);
        s_ExpHeap.Free(s_pDecodedPcmBufferR);
        s_ExpHeap.Free(s_pStateBuffer);

        // Finalize DSP and SND libraries
        NN_UTIL_PANIC_IF_FAILED(nn::snd::Finalize());
        NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
        nn::dsp::Finalize();
    }
}

void StartDemo()
{
    // Initialize fs
    nn::fs::Initialize();
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    // Initialize the heap
    s_ExpHeap.Initialize(
        nn::os::GetDeviceMemoryAddress(),
        DEVICE_MEMORY_SIZE
    );

    // Prepare RenderSystem
    uptr heapForGx = reinterpret_cast<uptr>(s_ExpHeap.Allocate(HEAP_SIZE_FOR_GX));
    NN_NULL_ASSERT(heapForGx);
    s_RenderSystem.Initialize(heapForGx, HEAP_SIZE_FOR_GX);

    // Create sound thread
    s_SoundThread.StartUsingAutoStack(
        SoundThreadFunc,
        NULL,
        SOUND_THREAD_STACK_SIZE,
        SOUND_THREAD_PRIORITY
    );

    // Create decoder thread
    s_DecodeThread.StartUsingAutoStack(
        DecodeThreadFunc,
        NULL,
        DECODE_THREAD_STACK_SIZE,
        DECODE_THREAD_PRIORITY
    );

    // After this, the application handles sleep itself
    TransitionHandler::EnableSleep();

    while(true)
    {
        // Render
        s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
        s_RenderSystem.Clear();
        s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
        s_RenderSystem.DrawText(10, 10, "CTR SampleDemos - aacdec/AacDecoder");
        s_RenderSystem.SwapBuffers();
        s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

        // Run sleep, HOME Menu, and POWER Menu transitions.
        TransitionHandler::Process();
        // Determine whether there has been an exit notification.
        if (TransitionHandler::IsExitRequired())
        {
            break;
        }
    }

    // Destroy decoder thread
    s_DecodeThreadExitEvent.Signal();
    s_DecodeThread.Join();
    s_DecodeThread.Finalize();

    // Destroy the sound thread
    s_IsSoundThreadEnabled = false;
    s_SoundThread.Join();
    s_SoundThread.Finalize();
    s_DecodeEvent.Finalize();
    s_DecodeThreadExitEvent.Finalize();

    s_ExpHeap.Free(reinterpret_cast<void*>(heapForGx));

    // Free contiguous region
    s_ExpHeap.Finalize();
}

void nnMain()
{
    // Initialize via the applet.
    // Sleep Mode is rejected automatically until TransitionHandler::EnableSleep is called.
    TransitionHandler::Initialize();
    // Here we need to check for exit notifications.
    if (!TransitionHandler::IsExitRequired())
    {
        StartDemo();
    }
    // Finalize via the applet.
    TransitionHandler::Finalize();

    // Finalize the application. The call to nn::applet::CloseApplication does not return.
    nn::applet::CloseApplication();
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
