﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     main.cpp

  Copyright (C)2009-2013 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 50863 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/fnd.h>
#include <nn/fs.h>
#include <nn/dsp.h>
#include <nn/applet.h>
#include <nn/aacenc.h>
#include <nn/hio.h>

#include <string>
#include "WaveFileReader.h"

extern const u8 DSPAUDIO_BEGIN[];

namespace
{
    nn::fnd::ExpHeap s_ExpHeap;
    const size_t DEVICE_MEMORY_SIZE = (32 * 1024 * 1024);

    const std::string INPUT_FILE_NAME = "rom:/fanfare_16bit_fs32kHz_1ch.wav";

    // Encoding parameter
    typedef struct
    {
        nn::aacenc::Rate bitrate;
        nn::aacenc::Form form;
    } EncodeParameter;
    const EncodeParameter c_EncodeParameter[] =
    {
        { nn::aacenc::RATE_L,   nn::aacenc::FORM_RAW,     },
        { nn::aacenc::RATE_L,   nn::aacenc::FORM_CRC_OFF, },
        { nn::aacenc::RATE_L,   nn::aacenc::FORM_CRC_ON,  },
        { nn::aacenc::RATE_M_L, nn::aacenc::FORM_RAW,     },
        { nn::aacenc::RATE_M_L, nn::aacenc::FORM_CRC_OFF, },
        { nn::aacenc::RATE_M_L, nn::aacenc::FORM_CRC_ON,  },
        { nn::aacenc::RATE_M_H, nn::aacenc::FORM_RAW,     },
        { nn::aacenc::RATE_M_H, nn::aacenc::FORM_CRC_OFF, },
        { nn::aacenc::RATE_M_H, nn::aacenc::FORM_CRC_ON,  },
        { nn::aacenc::RATE_H,   nn::aacenc::FORM_RAW,     },
        { nn::aacenc::RATE_H,   nn::aacenc::FORM_CRC_OFF, },
        { nn::aacenc::RATE_H,   nn::aacenc::FORM_CRC_ON,  },
    };

    const std::string& GetBitrateString(s32 idx)
    {
        static const std::string bitrate[4] = { "br0", "br1", "br2", "br3", };
        NN_ASSERT(idx >= 0 && idx <= 3);
        return bitrate[idx];
    }
    const std::string& GetFormString(s32 idx)
    {
        static const std::string form[3] = { "raw", "adts", "adtscrc", };
        NN_ASSERT(idx >= 0 && idx <= 2);
        return form[idx];
    }
    nn::aacenc::Fs GetFsIndex(s32 fs)
    {
        switch (fs)
        {
        case 48000: return nn::aacenc::FS_48;
        case 44100: return nn::aacenc::FS_44;
        case 32000: return nn::aacenc::FS_32;
        case 24000: return nn::aacenc::FS_24;
        case 22050: return nn::aacenc::FS_22;
        case 16000: return nn::aacenc::FS_16;
        case 12000: return nn::aacenc::FS_12;
        case 11025: return nn::aacenc::FS_11;
        case  8000: return nn::aacenc::FS_08;
        default   : return nn::aacenc::FS_UNKNOWN;
        }
    }
}

void nnMain()
{
    NN_LOG("Demo Start\n");

    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    // Initialize fs
    nn::fs::Initialize();
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    // Initialize dsp
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadComponent(DSPAUDIO_BEGIN, NN_DSPCOMPONENT_SIZE));

    // Initialize HostIo
    nn::hio::Initialize(reinterpret_cast<void*>(nn::os::GetDeviceMemoryAddress()));

    // Initialize the heap
    s_ExpHeap.Initialize(
        nn::os::GetDeviceMemoryAddress() + nn::hio::WORKMEMORY_SIZE,
        DEVICE_MEMORY_SIZE
    );

    for (int pidx = 0; pidx < sizeof(c_EncodeParameter) / sizeof(c_EncodeParameter[0]); pidx++)
    {
        // Open the input file
        WaveFileReader reader(INPUT_FILE_NAME.c_str());
        NN_ASSERT(reader.GetQuantumBits() == 16);

        // Open the output file
        nn::hio::HostFile hostFile;
        std::string filePathOut = OUTPUT_DIRECTORY;
        filePathOut += INPUT_FILE_NAME.substr(5, INPUT_FILE_NAME.size() - 9);  // Remove "rom:/" and ".wav"
        filePathOut += "_" + GetBitrateString(c_EncodeParameter[pidx].bitrate);
        filePathOut += "_" + GetFormString(c_EncodeParameter[pidx].form);
        filePathOut += ".aac";
        NN_LOG("%s\n", filePathOut.c_str() );
        NN_UTIL_PANIC_IF_FAILED(
            hostFile.Open(
                filePathOut.c_str(),
                nn::hio::HostFile::ACCESS_MODE_READ_WRITE,
                nn::hio::HostFile::OPEN_DISP_CREATE_ALWAYS
            )
        );

        // Prepare a buffer for the wav data, and load the data
        u32 size = reader.GetDataSize();
        s16* pPcmBuffer = reinterpret_cast<s16*>(s_ExpHeap.Allocate(size));
        NN_ASSERT(pPcmBuffer);
        reader.ReadData(pPcmBuffer, size);

        s16* pPcmBufferL = reinterpret_cast<s16*>(s_ExpHeap.Allocate(NN_DSP_CODEC_AAC_PCM_SAMPLES * sizeof(s16), 32));
        NN_ASSERT(pPcmBufferL);
        s16* pPcmBufferR = reinterpret_cast<s16*>(s_ExpHeap.Allocate(NN_DSP_CODEC_AAC_PCM_SAMPLES * sizeof(s16), 32));
        NN_ASSERT(pPcmBufferR);

        // Prepare stream buffer
        u8* pStreamBuffer = reinterpret_cast<u8*>(s_ExpHeap.Allocate(NN_DSP_CODEC_AAC_BS_SIZE, 32));
        NN_ASSERT(pStreamBuffer);

        // Start encoding
        nn::aacenc::Fs fsidx = GetFsIndex(reader.GetSampleRate());
        NN_ASSERT(fsidx != nn::aacenc::FS_UNKNOWN);
        if (nn::aacenc::Initialize(reader.GetChannelCount(), fsidx, c_EncodeParameter[pidx].bitrate, c_EncodeParameter[pidx].form) == false)
        {
            NN_PANIC("error: init\n");
        }
        int sampleLength = size / (reader.GetChannelCount() * sizeof(s16));
        s16* pInput = pPcmBuffer;
        while (sampleLength > 0)
        {
            int n = nn::math::Min(NN_DSP_CODEC_AAC_PCM_SAMPLES, sampleLength);
            sampleLength -= n;
            for (int i = 0; i < n; i++)
            {
                pPcmBufferL[i] = *pInput++;
                if (reader.GetChannelCount() == 2)
                {
                    pPcmBufferR[i] = *pInput++;
                }
            }
            // If there is not enough sample data, pad with zeroes
            for (int i = n; i < NN_DSP_CODEC_AAC_PCM_SAMPLES; i++)
            {
                pPcmBufferL[i] = 0;
                pPcmBufferR[i] = 0;
            }
            nn::dsp::FlushDataCache(reinterpret_cast<uptr>(pPcmBufferL), NN_DSP_CODEC_AAC_PCM_SAMPLES * sizeof(s16));
            nn::dsp::FlushDataCache(reinterpret_cast<uptr>(pPcmBufferR), NN_DSP_CODEC_AAC_PCM_SAMPLES * sizeof(s16));
            nn::aacenc::Encode(pPcmBufferL, pPcmBufferR, pStreamBuffer);
            nn::aacenc::Wait();
            hostFile.Write(pStreamBuffer, nn::aacenc::GetStreamSize());
        }
        nn::aacenc::Finalize();

        // Free memory
        s_ExpHeap.Free(pPcmBuffer);
        s_ExpHeap.Free(pPcmBufferL);
        s_ExpHeap.Free(pPcmBufferR);
        s_ExpHeap.Free(pStreamBuffer);

        // Close the file
        reader.Finalize();
        hostFile.Close();
    }

    // Free contiguous region
    s_ExpHeap.Finalize();

    // HIO termination
    nn::hio::Finalize();

    // Dispose of dsp
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
    nn::dsp::Finalize();

    NN_LOG("Demo End\n");
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
